@@ -5,17 +5,26 @@ from datetime import datetime
 from functools import wraps
 from logging import getLogger
 import re
-from urllib import urlencode
+from six import iterkeys, string_types, integer_types, iteritems, PY3
+from six.moves import xrange
+
+try:
+    # PY3
+    from urllib.parse import urlencode, quote_plus
+except ImportError:
+    # PY2
+    from urllib import urlencode, quote_plus
 
 import requests
-# import either simplejson or the json module in Python >= 2.6
-from requests.compat import json
+import simplejson as json  # for use_decimal
+from simplejson import JSONDecodeError
 
 from pyelasticsearch.downtime import DowntimePronePool
 from pyelasticsearch.exceptions import (Timeout, ConnectionError,
                                         ElasticHttpError,
                                         InvalidJsonResponseError,
-                                        ElasticHttpNotFoundError)
+                                        ElasticHttpNotFoundError,
+                                        IndexAlreadyExistsError)
 
 DATETIME_REGEX = re.compile(
     r'^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})T'
@@ -31,7 +40,17 @@ def _add_es_kwarg_docs(params, method):
     purpose of letting the user know that they're safe to use--we won't be
     paving over them in the future for something pyelasticsearch-specific.
     """
+    def docs_for_kwarg(p):
+        return '\n        :arg %s: See the ES docs.' % p
+
     doc = method.__doc__
+
+    # Handle the case where there are no :arg declarations to key off:
+    if '\n        :arg' not in doc:
+        first_param, params = params[0], params[1:]
+        doc = doc.replace('\n        (Insert es_kwargs here.)',
+                          docs_for_kwarg(first_param))
+
     for p in params:
         if ('\n        :arg %s: ' % p) not in doc:
             # Find the last documented arg so we can put our generated docs
@@ -43,7 +62,7 @@ def _add_es_kwarg_docs(params, method):
                 re.MULTILINE | re.DOTALL).end()
 
             doc = ''.join([doc[:insertion_point],
-                           '\n        :arg %s: See the ES docs.' % p,
+                           docs_for_kwarg(p),
                            doc[insertion_point:]])
     method.__doc__ = doc
 
@@ -71,7 +90,7 @@ def es_kwargs(*args_to_convert):
             # Make kwargs the map of normal kwargs and query_params the map of
             # kwargs destined for query string params:
             query_params = {}
-            for k, v in kwargs.items():  # NOT iteritems; we mutate kwargs
+            for k in list(iterkeys(kwargs)):  # Make a copy; we mutate kwargs.
                 if k.startswith('es_'):
                     query_params[k[3:]] = kwargs.pop(k)
                 elif k in convertible_args:
@@ -85,9 +104,14 @@ class ElasticSearch(object):
     """
     An object which manages connections to elasticsearch and acts as a
     go-between for API calls to it
+
+    This object is thread-safe. You can create one instance and share it
+    among all threads.
     """
     def __init__(self, urls, timeout=60, max_retries=0, revival_delay=300):
         """
+        :arg urls: A URL or iterable of URLs of ES nodes. These are full URLs
+            with port numbers, like ``http://elasticsearch.example.com:9200``.
         :arg timeout: Number of seconds to wait for each request before raising
             Timeout
         :arg max_retries: How many other servers to try, in series, after a
@@ -95,7 +119,7 @@ class ElasticSearch(object):
         :arg revival_delay: Number of seconds for which to avoid a server after
             it times out or is uncontactable
         """
-        if isinstance(urls, basestring):
+        if isinstance(urls, string_types):
             urls = [urls]
         urls = [u.rstrip('/') for u in urls]
         self.servers = DowntimePronePool(urls, revival_delay)
@@ -108,11 +132,11 @@ class ElasticSearch(object):
 
         json_converter = self.from_python
 
-        class DateSavvyJsonEncoder(json.JSONEncoder):
+        class JsonEncoder(json.JSONEncoder):
             def default(self, value):
                 """Convert more Python data types to ES-understandable JSON."""
                 return json_converter(value)
-        self.json_encoder = DateSavvyJsonEncoder
+        self.json_encoder = JsonEncoder
 
     def _concat(self, items):
         """
@@ -124,7 +148,7 @@ class ElasticSearch(object):
         # TODO: Why strip out _all?
         if items is None:
             return ''
-        if isinstance(items, basestring):
+        if isinstance(items, string_types):
             items = [items]
         return ','.join(i for i in items if i != '_all')
 
@@ -132,12 +156,14 @@ class ElasticSearch(object):
     def _to_query(cls, obj):
         """Convert a native-Python object to a query string representation."""
         # Quick and dirty thus far
-        if isinstance(obj, basestring):
+        if isinstance(obj, string_types):
             return obj
         if isinstance(obj, bool):
             return 'true' if obj else 'false'
-        if isinstance(obj, (long, int, float)):
+        if isinstance(obj, integer_types):
             return str(obj)
+        if isinstance(obj, float):
+            return repr(obj)  # str loses precision.
         if isinstance(obj, (list, tuple)):
             return ','.join(cls._to_query(o) for o in obj)
         iso = _iso_datetime(obj)
@@ -146,15 +172,29 @@ class ElasticSearch(object):
         raise TypeError("_to_query() doesn't know how to represent %r in an ES"
                         " query string." % obj)
 
-    def _send_request(self,
-                      method,
-                      path_components,
-                      body='',
-                      query_params=None,
-                      encode_body=True):
+    def _join_path(self, path_components):
+        """Smush together the (string) path components, ignoring empty ones."""
+        path = '/'.join(quote_plus(str(p), '') for p in path_components if
+                        len(str(p)))
+
+        if not path.startswith('/'):
+            path = '/' + path
+        return path
+
+    def send_request(self,
+                     method,
+                     path_components,
+                     body='',
+                     query_params=None,
+                     encode_body=True):
         """
         Send an HTTP request to ES, and return the JSON-decoded response.
 
+        This is mostly an internal method, but it also comes in handy if you
+        need to use a brand new ES API that isn't yet explicitly supported by
+        pyelasticsearch, while still taking advantage of our connection pooling
+        and retrying.
+
         Retry the request on different servers if the first one is down and
         ``self.max_retries`` > 0.
 
@@ -162,24 +202,17 @@ class ElasticSearch(object):
         :arg path_components: An iterable of path components, to be joined by
             "/"
         :arg body: The request body
-        :arg query_params: A map of querystring param names to values or None
+        :arg query_params: A map of querystring param names to values or
+            ``None``
         :arg encode_body: Whether to encode the body of the request as JSON
         """
-        def join_path(path_components):
-            """Smush together the path components, ignoring empty ones."""
-            path = '/'.join(str(p) for p in path_components if p)
-            if not path.startswith('/'):
-                path = '/' + path
-            return path
-
-        path = join_path(path_components)
+        path = self._join_path(path_components)
         if query_params:
             path = '?'.join(
                 [path, urlencode(dict((k, self._to_query(v)) for k, v in
-                                      query_params.iteritems()))])
+                                      iteritems(query_params)))])
 
-        kwargs = ({'data': self._encode_json(body) if encode_body else body}
-                   if body else {})
+        request_body = self._encode_json(body) if encode_body else body
         req_method = getattr(self.session, method.lower())
 
         # We do our own retrying rather than using urllib3's; we want to retry
@@ -189,13 +222,14 @@ class ElasticSearch(object):
             server_url, was_dead = self.servers.get()
             url = server_url + path
             self.logger.debug(
-                'making %s request to path: %s %s with body: %s',
-                method, url, path, kwargs.get('data', {}))
+                "Making a request equivalent to this: curl -X%s '%s' -d '%s'" %
+                (method, url, request_body))
+
             try:
-                # prefetch=True so the connection can be quickly returned to
-                # the pool. This is the default in requests >=0.3.16.
                 resp = req_method(
-                    url, prefetch=True, timeout=self.timeout, **kwargs)
+                    url,
+                    timeout=self.timeout,
+                    **({'data': request_body} if body else {}))
             except (ConnectionError, Timeout):
                 self.servers.mark_dead(server_url)
                 self.logger.info('%s marked as dead for %s seconds.',
@@ -211,29 +245,38 @@ class ElasticSearch(object):
         self.logger.debug('response status: %s', resp.status_code)
         prepped_response = self._decode_response(resp)
         if resp.status_code >= 400:
-            error_class = (ElasticHttpNotFoundError if resp.status_code == 404
-                           else ElasticHttpError)
-            raise error_class(
-                resp.status_code,
-                prepped_response.get('error', prepped_response))
+            self._raise_exception(resp, prepped_response)
         self.logger.debug('got response %s', prepped_response)
         return prepped_response
 
+    def _raise_exception(self, response, decoded_body):
+        """Raise an exception based on an error-indicating response from ES."""
+        error_message = decoded_body.get('error', decoded_body)
+
+        error_class = ElasticHttpError
+        if response.status_code == 404:
+            error_class = ElasticHttpNotFoundError
+        elif error_message.startswith('IndexAlreadyExistsException'):
+            error_class = IndexAlreadyExistsError
+
+        raise error_class(response.status_code, error_message)
+
     def _encode_json(self, body):
         """Return body encoded as JSON."""
-        return json.dumps(body, cls=self.json_encoder)
+        return json.dumps(body, cls=self.json_encoder, use_decimal=True)
 
     def _decode_response(self, response):
         """Return a native-Python representation of a response's JSON blob."""
-        json_response = response.json
-        if json_response is None:
+        try:
+            json_response = response.json()
+        except JSONDecodeError:
             raise InvalidJsonResponseError(response)
         return json_response
 
     ## REST API
 
     @es_kwargs('routing', 'parent', 'timestamp', 'ttl', 'percolate',
-               'consistency', 'replication', 'refresh', 'timeout')
+               'consistency', 'replication', 'refresh', 'timeout', 'fields')
     def index(self, index, doc_type, doc, id=None, force_insert=False,
               query_params=None):
         """
@@ -241,7 +284,8 @@ class ElasticSearch(object):
 
         :arg index: The name of the index to which to add the document
         :arg doc_type: The type of the document
-        :arg doc: A mapping, convertible to JSON, representing the document
+        :arg doc: A Python mapping object, convertible to JSON, representing
+            the document
         :arg id: The ID to give the document. Leave blank to make one up.
         :arg force_insert: If ``True`` and a document of the given ID already
             exists, fail rather than updating it.
@@ -285,10 +329,10 @@ class ElasticSearch(object):
         if force_insert:
             query_params['op_type'] = 'create'
 
-        return self._send_request('POST' if id is None else 'PUT',
-                                  [index, doc_type, id],
-                                  doc,
-                                  query_params)
+        return self.send_request('POST' if id is None else 'PUT',
+                                 [index, doc_type, id],
+                                 doc,
+                                 query_params)
 
     @es_kwargs('consistency', 'refresh')
     def bulk_index(self, index, doc_type, docs, id_field='id',
@@ -298,8 +342,8 @@ class ElasticSearch(object):
 
         :arg index: The name of the index to which to add the document
         :arg doc_type: The type of the document
-        :arg docs: An iterable of mappings, convertible to JSON, representing
-            documents to index
+        :arg docs: An iterable of Python mapping objects, convertible to JSON,
+            representing documents to index
         :arg id_field: The field of each document that holds its ID
 
         See `ES's bulk API`_ for more detail.
@@ -324,11 +368,11 @@ class ElasticSearch(object):
         # Need the trailing newline.
         body = '\n'.join(body_bits) + '\n'
         query_params['op_type'] = 'create'  # TODO: Why?
-        return self._send_request('POST',
-                                  [index, '_bulk'],
-                                  body,
-                                  encode_body=False,
-                                  query_params=query_params)
+        return self.send_request('POST',
+                                 ['_bulk'],
+                                 body,
+                                 encode_body=False,
+                                 query_params=query_params)
 
     @es_kwargs('routing', 'parent', 'replication', 'consistency', 'refresh')
     def delete(self, index, doc_type, id, query_params=None):
@@ -337,16 +381,21 @@ class ElasticSearch(object):
 
         :arg index: The name of the index from which to delete
         :arg doc_type: The type of the document to delete
-        :arg id: The ID of the document to delete
+        :arg id: The (string or int) ID of the document to delete
 
         See `ES's delete API`_ for more detail.
 
         .. _`ES's delete API`:
             http://www.elasticsearch.org/guide/reference/api/delete.html
         """
-        # TODO: Raise ValueError if id boils down to a 0-length string.
-        return self._send_request('DELETE', [index, doc_type, id],
-                                  query_params=query_params)
+        # id should never be None, and it's not particular dangerous
+        # (equivalent to deleting a doc with ID "None", but it's almost
+        # certainly not what the caller meant:
+        if id is None or id == '':
+            raise ValueError('No ID specified. To delete all documents in '
+                             'an index, use delete_all().')
+        return self.send_request('DELETE', [index, doc_type, id],
+                                 query_params=query_params)
 
     @es_kwargs('routing', 'parent', 'replication', 'consistency', 'refresh')
     def delete_all(self, index, doc_type, query_params=None):
@@ -363,8 +412,8 @@ class ElasticSearch(object):
         .. _`ES's delete API`:
             http://www.elasticsearch.org/guide/reference/api/delete.html
         """
-        return self._send_request('DELETE', [index, doc_type],
-                                  query_params=query_params)
+        return self.send_request('DELETE', [index, doc_type],
+                                 query_params=query_params)
 
     @es_kwargs('q', 'df', 'analyzer', 'default_operator', 'source' 'routing',
                'replication', 'consistency')
@@ -381,8 +430,8 @@ class ElasticSearch(object):
         .. _`ES's delete-by-query API`:
             http://www.elasticsearch.org/guide/reference/api/delete-by-query.html
         """
-        return self._send_request('DELETE', [index, doc_type, '_query'], query,
-                                  query_params=query_params)
+        return self.send_request('DELETE', [index, doc_type, '_query'], query,
+                                 query_params=query_params)
 
     @es_kwargs('realtime', 'fields', 'routing', 'preference', 'refresh')
     def get(self, index, doc_type, id, query_params=None):
@@ -398,24 +447,65 @@ class ElasticSearch(object):
         .. _`ES's get API`:
             http://www.elasticsearch.org/guide/reference/api/get.html
         """
-        return self._send_request('GET', [index, doc_type, id],
-                                  query_params=query_params)
+        return self.send_request('GET', [index, doc_type, id],
+                                 query_params=query_params)
+
+
+    @es_kwargs('routing', 'parent', 'timeout', 'replication', 'consistency',
+               'percolate', 'refresh', 'retry_on_conflict', 'fields')
+    def update(self, index, doc_type, id, script=None, params=None, lang=None,
+               query_params=None, doc=None, upsert=None):
+        """
+        Update an existing document. Raise ``TypeError`` if ``script``, ``doc``
+        and ``upsert`` are all unspecified.
+
+        :arg index: The name of the index containing the document
+        :arg doc_type: The type of the document
+        :arg id: The ID of the document
+        :arg script: The script to be used to update the document
+        :arg params: A dict of the params to be put in scope of the script
+        :arg lang: The language of the script. Omit to use the default,
+            specified by ``script.default_lang``.
+        :arg doc: A partial document to be merged into the existing document
+        :arg upsert: The content for the new document created if the document
+            does not exist
+        """
+        if script is None and doc is None and upsert is None:
+            raise TypeError('At least one of the script, doc, or upsert '
+                            'kwargs must be provided.')
+
+        body = {}
+        if script:
+            body['script'] = script
+        if lang and script:
+            body['lang'] = lang
+        if doc:
+            body['doc'] = doc
+        if upsert:
+            body['upsert'] = upsert
+        if params:
+            body['params'] = params
+        return self.send_request(
+            'POST',
+            [index, doc_type, id, '_update'],
+            body=body,
+            query_params=query_params)
 
     def _search_or_count(self, kind, query, index=None, doc_type=None,
                          query_params=None):
-        if isinstance(query, basestring):
+        if isinstance(query, string_types):
             query_params['q'] = query
             body = ''
         else:
             body = query
 
-        return self._send_request(
+        return self.send_request(
             'GET',
             [self._concat(index), self._concat(doc_type), kind],
             body,
             query_params=query_params)
 
-    @es_kwargs('routing')
+    @es_kwargs('routing', 'size')
     def search(self, query, **kwargs):
         """
         Execute a search query against one or more indices and get back search
@@ -424,8 +514,12 @@ class ElasticSearch(object):
         :arg query: A dictionary that will convert to ES's query DSL or a
             string that will serve as a textual query to be passed as the ``q``
             query string parameter
-        :arg index: An index or iterable of indexes to search
-        :arg doc_type: A document type or iterable thereof to search
+        :arg index: An index or iterable of indexes to search. Omit to search
+            all.
+        :arg doc_type: A document type or iterable thereof to search. Omit to
+            search all.
+        :arg size: Limit the number of results to ``size``. Use with ``es_from`` to
+            implement paginated searching.
 
         See `ES's search API`_ for more detail.
 
@@ -442,8 +536,10 @@ class ElasticSearch(object):
         :arg query: A dictionary that will convert to ES's query DSL or a
             string that will serve as a textual query to be passed as the ``q``
             query string parameter
-        :arg index: An index or iterable of indexes to search
-        :arg doc_type: A document type or iterable thereof to search
+        :arg index: An index or iterable of indexes to search. Omit to search
+            all.
+        :arg doc_type: A document type or iterable thereof to search. Omit to
+            search all.
 
         See `ES's count API`_ for more detail.
 
@@ -469,7 +565,7 @@ class ElasticSearch(object):
         """
         # TODO: Think about turning index=None into _all if doc_type is non-
         # None, per the ES doc page.
-        return self._send_request(
+        return self.send_request(
             'GET',
             [self._concat(index), self._concat(doc_type), '_mapping'],
             query_params=query_params)
@@ -494,19 +590,19 @@ class ElasticSearch(object):
         # don't need to expose the "_all" magic string. We haven't done it yet
         # since this routine is not dangerous: ES makes you explicily pass
         # "_all" to update all mappings.
-        return self._send_request(
+        return self.send_request(
             'PUT',
             [self._concat(index), doc_type, '_mapping'],
             mapping,
             query_params=query_params)
 
     @es_kwargs('search_type', 'search_indices', 'search_types',
-               'search_scroll', 'search_size', 'search_from', 'fields',
+               'search_scroll', 'search_size', 'search_from',
                'like_text', 'percent_terms_to_match', 'min_term_freq',
                'max_query_terms', 'stop_words', 'min_doc_freq', 'max_doc_freq',
                'min_word_len', 'max_word_len', 'boost_terms', 'boost',
                'analyzer')
-    def more_like_this(self, index, doc_type, id, fields, query_params=None):
+    def more_like_this(self, index, doc_type, id, mlt_fields, body='', query_params=None):
         """
         Execute a "more like this" search query against one or more fields and
         get back search hits.
@@ -515,17 +611,20 @@ class ElasticSearch(object):
             lives
         :arg doc_type: The type of document to find others like
         :arg id: The ID of the document to find others like
-        :arg fields: A list of fields to compare on
+        :arg mlt_fields: The list of fields to compare on
+        :arg body: A dictionary that will convert to ES's query DSL and be
+            passed as the request body
 
         See `ES's more-like-this API`_ for more detail.
 
         .. _`ES's more-like-this API`:
             http://www.elasticsearch.org/guide/reference/api/more-like-this.html
         """
-        query_params['fields'] = self._concat(fields)  # TODO: ES docs say "mlt_fields".
-        return self._send_request('GET',
-                                  [index, doc_type, id, '_mlt'],
-                                  query_params=query_params)
+        query_params['mlt_fields'] = self._concat(mlt_fields)
+        return self.send_request('GET',
+                                 [index, doc_type, id, '_mlt'],
+                                 body=body,
+                                 query_params=query_params)
 
     ## Index Admin API
 
@@ -541,8 +640,38 @@ class ElasticSearch(object):
         .. _`ES's index-status API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-status.html
         """
-        return self._send_request('GET', [self._concat(index), '_status'],
-                                  query_params=query_params)
+        return self.send_request('GET', [self._concat(index), '_status'],
+                                 query_params=query_params)
+
+    @es_kwargs()
+    def update_aliases(self, settings, query_params=None):
+        """
+        Add, remove, or update aliases in bulk.
+
+        :arg settings: a dictionary specifying the actions to perform
+
+        See `ES's admin-indices-aliases API`_.
+
+        .. _`ES's admin-indices-aliases API`:
+            http://www.elasticsearch.org/guide/reference/api/admin-indices-aliases.html
+        """
+        return self.send_request('POST', ['_aliases'],
+                                 body=settings, query_params=query_params)
+
+    @es_kwargs()
+    def aliases(self, index=None, query_params=None):
+        """
+        Retrieve a listing of aliases
+
+        :arg index: the name of an index or an iterable of indices
+
+        See `ES's admin-indices-aliases API`_.
+
+        .. _`ES's admin-indices-aliases API`:
+            http://www.elasticsearch.org/guide/reference/api/admin-indices-aliases.html
+        """
+        return self.send_request('GET', [self._concat(index), '_aliases'],
+                                 query_params=query_params)
 
     @es_kwargs()
     def create_index(self, index, settings=None, query_params=None):
@@ -552,13 +681,16 @@ class ElasticSearch(object):
         :arg index: The name of the index to create
         :arg settings: A dictionary of settings
 
+        If the index already exists, raise
+        :class:`~pyelasticsearch.exceptions.IndexAlreadyExistsError`.
+
         See `ES's create-index API`_ for more detail.
 
         .. _`ES's create-index API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-create-index.html
         """
-        return self._send_request('PUT', [index], body=settings,
-                                  query_params=query_params)
+        return self.send_request('PUT', [index], body=settings,
+                                 query_params=query_params)
 
     @es_kwargs()
     def delete_index(self, index, query_params=None):
@@ -567,6 +699,9 @@ class ElasticSearch(object):
 
         :arg index: An index or iterable thereof to delete
 
+        If the index is not found, raise
+        :class:`~pyelasticsearch.exceptions.ElasticHttpNotFoundError`.
+
         See `ES's delete-index API`_ for more detail.
 
         .. _`ES's delete-index API`:
@@ -575,8 +710,8 @@ class ElasticSearch(object):
         if not index:
             raise ValueError('No indexes specified. To delete all indexes, use'
                              ' delete_all_indexes().')
-        return self._send_request('DELETE', [self._concat(index)],
-                                  query_params=query_params)
+        return self.send_request('DELETE', [self._concat(index)],
+                                 query_params=query_params)
 
     def delete_all_indexes(self, **kwargs):
         """Delete all indexes."""
@@ -594,8 +729,8 @@ class ElasticSearch(object):
         .. _`ES's close-index API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-open-close.html
         """
-        return self._send_request('POST', [index, '_close'],
-                                  query_params=query_params)
+        return self.send_request('POST', [index, '_close'],
+                                 query_params=query_params)
 
     @es_kwargs()
     def open_index(self, index, query_params=None):
@@ -609,8 +744,24 @@ class ElasticSearch(object):
         .. _`ES's open-index API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-open-close.html
         """
-        return self._send_request('POST', [index, '_open'],
-                                  query_params=query_params)
+        return self.send_request('POST', [index, '_open'],
+                                 query_params=query_params)
+
+    @es_kwargs()
+    def get_settings(self, index, query_params=None):
+        """
+        Get the settings of one or more indexes.
+
+        :arg index: An index or iterable of indexes
+
+        See `ES's get-settings API`_ for more detail.
+
+        .. _`ES's get-settings API`:
+            http://www.elasticsearch.org/guide/reference/api/admin-indices-get-settings.html
+        """
+        return self.send_request('GET',
+                                 [self._concat(index), '_settings'],
+                                 query_params=query_params)
 
     @es_kwargs()
     def update_settings(self, index, settings, query_params=None):
@@ -630,10 +781,10 @@ class ElasticSearch(object):
                              ' update_all_settings().')
         # If we implement the "update cluster settings" API, call that
         # update_cluster_settings().
-        return self._send_request('PUT',
-                                  [self._concat(index), '_settings'],
-                                  body=settings,
-                                  query_params=query_params)
+        return self.send_request('PUT',
+                                [self._concat(index), '_settings'],
+                                body=settings,
+                                query_params=query_params)
 
     @es_kwargs()
     def update_all_settings(self, settings, query_params=None):
@@ -647,8 +798,8 @@ class ElasticSearch(object):
         .. _`ES's update-settings API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-update-settings.html
         """
-        return self._send_request('PUT', ['_settings'], body=settings,
-                                  query_params=query_params)
+        return self.send_request('PUT', ['_settings'], body=settings,
+                                 query_params=query_params)
 
     @es_kwargs('refresh')
     def flush(self, index=None, query_params=None):
@@ -662,9 +813,9 @@ class ElasticSearch(object):
         .. _`ES's flush API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-flush.html
         """
-        return self._send_request('POST',
-                                  [self._concat(index), '_flush'],
-                                  query_params=query_params)
+        return self.send_request('POST',
+                                 [self._concat(index), '_flush'],
+                                 query_params=query_params)
 
     @es_kwargs()
     def refresh(self, index=None, query_params=None):
@@ -678,8 +829,8 @@ class ElasticSearch(object):
         .. _`ES's refresh API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-refresh.html
         """
-        return self._send_request('POST', [self._concat(index), '_refresh'],
-                                  query_params=query_params)
+        return self.send_request('POST', [self._concat(index), '_refresh'],
+                                 query_params=query_params)
 
     @es_kwargs()
     def gateway_snapshot(self, index=None, query_params=None):
@@ -693,7 +844,7 @@ class ElasticSearch(object):
         .. _`ES's gateway-snapshot API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-gateway-snapshot.html
         """
-        return self._send_request(
+        return self.send_request(
             'POST',
             [self._concat(index), '_gateway', 'snapshot'],
             query_params=query_params)
@@ -711,9 +862,9 @@ class ElasticSearch(object):
         .. _`ES's optimize API`:
             http://www.elasticsearch.org/guide/reference/api/admin-indices-optimize.html
         """
-        return self._send_request('POST',
-                                  [self._concat(index), '_optimize'],
-                                  query_params=query_params)
+        return self.send_request('POST',
+                                 [self._concat(index), '_optimize'],
+                                 query_params=query_params)
 
     @es_kwargs('level', 'wait_for_status', 'wait_for_relocating_shards',
                'wait_for_nodes', 'timeout')
@@ -728,11 +879,28 @@ class ElasticSearch(object):
         .. _`ES's cluster-health API`:
             http://www.elasticsearch.org/guide/reference/api/admin-cluster-health.html
         """
-        return self._send_request(
+        return self.send_request(
             'GET',
             ['_cluster', 'health', self._concat(index)],
             query_params=query_params)
 
+    @es_kwargs('filter_nodes', 'filter_routing_table', 'filter_metadata',
+               'filter_blocks', 'filter_indices')
+    def cluster_state(self, query_params=None):
+        """
+        The cluster state API allows to get comprehensive state
+        information of the whole cluster.
+
+        (Insert es_kwargs here.)
+
+        See `ES's cluster-state API`_ for more detail.
+
+        .. _`ES's cluster-state API`:
+            http://www.elasticsearch.org/guide/reference/api/admin-cluster-state.html
+        """
+        return self.send_request(
+            'GET', ['_cluster', 'state'], query_params=query_params)
+
     def from_python(self, value):
         """
         Convert Python values to a form suitable for ElasticSearch's JSON.
@@ -740,43 +908,10 @@ class ElasticSearch(object):
         iso = _iso_datetime(value)
         if iso:
             return iso
-        if isinstance(value, str):
+        if not PY3 and isinstance(value, str):
             return unicode(value, errors='replace')  # TODO: Be stricter.
-        return value
-
-    def to_python(self, value):
-        """Convert values from ElasticSearch to native Python values."""
-        if isinstance(value, (int, float, long, complex, list, tuple, bool)):
-            return value
-
-        if isinstance(value, basestring):
-            possible_datetime = DATETIME_REGEX.search(value)
-
-            if possible_datetime:
-                date_values = possible_datetime.groupdict()
-
-                for dk, dv in date_values.items():
-                    date_values[dk] = int(dv)
-
-                return datetime(
-                    date_values['year'], date_values['month'],
-                    date_values['day'], date_values['hour'],
-                    date_values['minute'], date_values['second'])
-
-        try:
-            # This is slightly gross but it's hard to tell otherwise what the
-            # string's original type might have been. Be careful who you trust.
-            converted_value = eval(value)
-
-            # Try to handle most built-in types.
-            if isinstance(
-                    converted_value,
-                    (list, tuple, set, dict, int, float, long, complex)):
-                return converted_value
-        except Exception:
-            # If it fails (SyntaxError or its ilk) or we don't trust it,
-            # continue on.
-            pass
+        if isinstance(value, set):
+            return list(value)
 
         return value
 
