@@ -2,8 +2,11 @@
 """
 Unit tests for pyelasticsearch.  These require an elasticsearch server running on the default port (localhost:9200).
 """
+import sys
 from datetime import datetime, date
+from decimal import Decimal
 import unittest
+import six
 
 from mock import patch
 import requests
@@ -15,7 +18,8 @@ from pyelasticsearch.client import es_kwargs
 
 def arbitrary_response():
     response = requests.Response()
-    response._content = '{"some": "json"}'
+    response._content = six.b('{"some": "json"}')
+    response.status_code = 200
     return response
 
 
@@ -26,7 +30,7 @@ class ElasticSearchTestCase(unittest.TestCase):
     def tearDown(self):
         try:
             self.conn.delete_index('test-index')
-        except:
+        except Exception:
             pass
 
     def assertResultContains(self, result, expected):
@@ -39,6 +43,14 @@ class IndexingTestCase(ElasticSearchTestCase):
         result = self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id=1)
         self.assertResultContains(result, {'_type': 'test-type', '_id': '1', 'ok': True, '_index': 'test-index'})
 
+    def testIndexingWith0ID(self):
+        result = self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id=0)
+        self.assertResultContains(result, {'_type': 'test-type', '_id': '0', 'ok': True, '_index': 'test-index'})
+
+    def testQuotedCharsInID(self):
+        result = self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id="""<>?,./`~!@#$%^&*()_+=[]\{{}|:";'""")
+        self.assertResultContains(result, {'_type': 'test-type', '_id': """<>?,./`~!@#$%^&*()_+=[]\{{}|:";'""", 'ok': True, '_index': 'test-index'})
+
     def testIndexingWithoutID(self):
         result = self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'})
         self.assertResultContains(result, {'_type': 'test-type', 'ok': True, '_index': 'test-index'})
@@ -62,39 +74,64 @@ class IndexingTestCase(ElasticSearchTestCase):
         result = self.conn.open_index('test-index')
         self.assertResultContains(result, {'acknowledged': True, 'ok': True})
 
+    def testGetSettings(self):
+        self.conn.create_index('test-index')
+        result = self.conn.get_settings('test-index')
+        self.assertTrue('test-index'in result)
+        self.assertTrue('settings' in result['test-index'])
+
     def testUpdateSettings(self):
         """Make sure ``update_settings()`` sends the expected request."""
-        with patch.object(self.conn, '_send_request') as _send_request:
+        with patch.object(self.conn, 'send_request') as send_request:
             self.conn.update_settings(['test-index', 'toast-index'],
                                       {'index': {'number_of_replicas': 2}})
-        _send_request.assert_called_once_with(
+        send_request.assert_called_once_with(
             'PUT',
             ['test-index,toast-index', '_settings'],
             body={'index': {'number_of_replicas': 2}},
             query_params={})
 
     def testHealth(self):
-        with patch.object(self.conn, '_send_request') as _send_request:
+        with patch.object(self.conn, 'send_request') as send_request:
             self.conn.health(['test-index', 'toast-index'],
                              wait_for_status='yellow',
                              wait_for_nodes='>=1')
-        _send_request.assert_called_once_with(
+        send_request.assert_called_once_with(
             'GET',
             ['_cluster', 'health', 'test-index,toast-index'],
             query_params={'wait_for_status': 'yellow',
                           'wait_for_nodes': '>=1'})
 
-        with patch.object(self.conn, '_send_request') as _send_request:
+        with patch.object(self.conn, 'send_request') as send_request:
             self.conn.health()
-        _send_request.assert_called_once_with(
+        send_request.assert_called_once_with(
             'GET', ['_cluster', 'health', ''], query_params={})
 
+    def testClusterState(self):
+        result = self.conn.cluster_state(filter_routing_table=True)
+        self.assertTrue('nodes' in result)
+        self.assertFalse('routing_table' in result)
+
     def testDeleteByID(self):
         self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id=1)
         self.conn.refresh(['test-index'])
         result = self.conn.delete('test-index', 'test-type', 1)
         self.assertResultContains(result, {'_type': 'test-type', '_id': '1', 'ok': True, '_index': 'test-index'})
 
+    def testDeleteBy0ID(self):
+        self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id=0)
+        self.conn.refresh(['test-index'])
+        result = self.conn.delete('test-index', 'test-type', 0)
+        self.assertResultContains(result, {'_type': 'test-type', '_id': '0', 'ok': True, '_index': 'test-index'})
+
+    def testDeleteByIdWithoutId(self):
+        self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id=1)
+        self.conn.refresh(['test-index'])
+        self.assertRaises(
+            ValueError, self.conn.delete, 'test-index', 'test-type', '')
+        self.assertRaises(
+            ValueError, self.conn.delete, 'test-index', 'test-type', None)
+
     def testDeleteByDocType(self):
         self.conn.index('test-index', 'test-type', {'name': 'Joe Tester'}, id=1)
         self.conn.refresh(["test-index"])
@@ -133,7 +170,8 @@ class IndexingTestCase(ElasticSearchTestCase):
 
     def testCannotCreateExistingIndex(self):
         self.conn.create_index('another-index')
-        self.assertRaises(ElasticHttpError, self.conn.create_index, 'another-index')
+        self.assertRaises(
+            IndexAlreadyExistsError, self.conn.create_index, 'another-index')
         self.conn.delete_index('another-index')
         self.assertRaises(ElasticHttpError, self.conn.delete_index, 'another-index')
 
@@ -175,32 +213,6 @@ class IndexingTestCase(ElasticSearchTestCase):
         self.conn.delete_index('another-index')
         self.assertResultContains(result, {'ok': True})
 
-    def testFromPython(self):
-        self.assertEqual(self.conn.from_python('abc'), u'abc')
-        self.assertEqual(self.conn.from_python(u'☃'), u'☃')
-        self.assertEqual(self.conn.from_python(123), 123)
-        self.assertEqual(self.conn.from_python(12.2), 12.2)
-        self.assertEqual(self.conn.from_python(True), True)
-        self.assertEqual(self.conn.from_python(False), False)
-        self.assertEqual(self.conn.from_python(date(2011, 12, 30)), '2011-12-30T00:00:00')
-        self.assertEqual(self.conn.from_python(datetime(2011, 12, 30, 11, 59, 32)), '2011-12-30T11:59:32')
-        self.assertEqual(self.conn.from_python([1, 2, 3]), [1, 2, 3])
-        self.assertEqual(self.conn.from_python(set(['a', 'b', 'c'])), set(['a', 'b', 'c']))
-        self.assertEqual(self.conn.from_python({'a': 1, 'b': 3, 'c': 2}), {'a': 1, 'b': 3, 'c': 2})
-
-    def testToPython(self):
-        self.assertEqual(self.conn.to_python(u'abc'), u'abc')
-        self.assertEqual(self.conn.to_python(u'☃'), u'☃')
-        self.assertEqual(self.conn.to_python(123), 123)
-        self.assertEqual(self.conn.to_python(12.2), 12.2)
-        self.assertEqual(self.conn.to_python(True), True)
-        self.assertEqual(self.conn.to_python(False), False)
-        self.assertEqual(self.conn.to_python('2011-12-30T00:00:00'), datetime(2011, 12, 30))
-        self.assertEqual(self.conn.to_python('2011-12-30T11:59:32'), datetime(2011, 12, 30, 11, 59, 32))
-        self.assertEqual(self.conn.to_python([1, 2, 3]), [1, 2, 3])
-        self.assertEqual(self.conn.to_python(set(['a', 'b', 'c'])), set(['a', 'b', 'c']))
-        self.assertEqual(self.conn.to_python({'a': 1, 'b': 3, 'c': 2}), {'a': 1, 'b': 3, 'c': 2})
-
     def testBulkIndex(self):
         # Try counting the docs in a nonexistent index:
         self.assertRaises(ElasticHttpError, self.conn.count, '*:*', index=['test-index'])
@@ -220,15 +232,78 @@ class IndexingTestCase(ElasticSearchTestCase):
 
     def testErrorHandling(self):
         # Wrong port.
-        conn = ElasticSearch('http://example.com:1009200/')
+        conn = ElasticSearch('http://localhost:1009200/')
         self.assertRaises(ConnectionError, conn.count, '*:*')
 
         # Test invalid JSON.
         self.assertRaises(ValueError, conn._encode_json, object())
         resp = requests.Response()
-        resp._content = '{"busted" "json" "that": ["is] " wrong'
+        resp._content = six.b('{"busted" "json" "that": ["is] " wrong')
         self.assertRaises(InvalidJsonResponseError, conn._decode_response, resp)
 
+    def testUpdate(self):
+        """Smoke-test the ``update()`` API."""
+        SCRIPT = 'ctx._source.thing += count'
+        with patch.object(self.conn, 'send_request') as send_request:
+            self.conn.update('some_index',
+                             'some_type',
+                             3,
+                             SCRIPT,
+                             params={'count': 5},
+                             lang='python')
+        send_request.assert_called_once_with(
+            'POST', ['some_index', 'some_type', 3, '_update'],
+            body={'script': SCRIPT,
+                  'params': {'count': 5},
+                  'lang': 'python'},
+                  query_params={})
+
+    def testAliasIndex(self):
+        self.conn.create_index('test-index')
+        settings = {
+            "actions": [
+                {"add": {"index": "test-index", "alias": "test-alias"}}
+            ]
+        }
+        result = self.conn.update_aliases(settings)
+        self.assertResultContains(result, {'acknowledged': True, 'ok': True})
+
+    def testAliasNonexistentIndex(self):
+        settings = {
+            "actions": [
+                {"add": {"index": "test1", "alias": "alias1"}}
+            ]
+        }
+        self.assertRaises(ElasticHttpNotFoundError,
+                          self.conn.update_aliases,
+                          settings)
+
+    def testListAliases(self):
+        self.conn.create_index('test-index')
+        settings = {
+            "actions": [
+                {"add": {"index": "test-index", "alias": "test-alias"}}
+            ]
+        }
+        self.conn.update_aliases(settings)
+        result = self.conn.aliases('test-index')
+        self.assertEqual(result, {u'test-index': {u'aliases': {u'test-alias': {}}}})
+
+    def test_empty_path_segments(self):
+        """'' segments passed to ``_join_path`` should be omitted."""
+        # Call _join_path like get_mapping might if called with no params:
+        self.assertEqual(self.conn._join_path(['', '', '_mapping']),
+                         '/_mapping')
+
+    def test_0_path_segments(self):
+        """
+        ``0`` segments passed to ``_join_path`` should be included.
+
+        This is so doc IDs that are 0 work.
+        """
+        self.assertEqual(self.conn._join_path([0, '_mapping']),
+                         '/0/_mapping')
+
 
 class SearchTestCase(ElasticSearchTestCase):
     def setUp(self):
@@ -249,6 +324,16 @@ class SearchTestCase(ElasticSearchTestCase):
         result = self.conn.search('name:joe', index='test-index')
         self.assertResultContains(result, {'hits': {'hits': [{'_score': 0.19178301, '_type': 'test-type', '_id': '1', '_source': {'name': 'Joe Tester'}, '_index': 'test-index'}], 'total': 1, 'max_score': 0.19178301}})
 
+    def testSearchStringPaginated(self):
+        with patch.object(self.conn, 'send_request') as send_request:
+            self.conn.search('*:*', index='test-index', es_from=1, size=1)
+
+        send_request.assert_called_once_with(
+            'GET',
+            ['test-index', '', '_search'],
+            '',
+            query_params={'q': '*:*', 'from': 1, 'size': 1})
+
     def testSearchByDSL(self):
         self.conn.index('test-index', 'test-type', {'name': 'AgeJoe Tester', 'age': 25}, id=1)
         self.conn.index('test-index', 'test-type', {'name': 'AgeBill Baloney', 'age': 35}, id=2)
@@ -279,18 +364,50 @@ class SearchTestCase(ElasticSearchTestCase):
         result = self.conn.more_like_this('test-index', 'test-type', 1, ['name'], min_term_freq=1, min_doc_freq=1)
         self.assertResultContains(result, {'hits': {'hits': [{'_score': 0.19178301, '_type': 'test-type', '_id': '3', '_source': {'name': 'Joe Test'}, '_index': 'test-index'}], 'total': 1, 'max_score': 0.19178301}})
 
+    def testMLTWithBody(self):
+        self.conn.index('test-index', 'test-type', {'name': 'Joe Test', 'age': 22}, id=2)
+        self.conn.index('test-index', 'test-type', {'name': 'Joe Justin', 'age': 16}, id=3)
+        self.conn.refresh(['test-index'])
+
+        body = {'filter': {
+                    'fquery': {
+                        'query': {
+                            'range': {
+                                'age': {
+                                    'from': 10,
+                                    'to': 20
+                                },
+                            },
+                        }
+                    }
+                }
+            }
+        result = self.conn.more_like_this('test-index', 'test-type', 1, ['name'], body=body, min_term_freq=1, min_doc_freq=1)
+        self.assertResultContains(result,
+                {'hits': {'hits': [{'_score': 0.19178301, '_type': 'test-type', '_id': '3', '_source': {'age': 16, 'name': 'Joe Justin'}, '_index': 'test-index'}], 'total': 1, 'max_score': 0.19178301}})
+
+    def testMLTFields(self):
+        self.conn.index('test-index', 'test-type', {'name': 'Angus', 'sport': 'football'}, id=3)
+        self.conn.index('test-index', 'test-type', {'name': 'Cam', 'sport': 'football'}, id=4)
+        self.conn.index('test-index', 'test-type', {'name': 'Sophia', 'sport': 'baseball'}, id=5)
+
+        self.conn.refresh(['test-index'])
+
+        result = self.conn.more_like_this('test-index', 'test-type', 3, ['sport'], min_term_freq=1, min_doc_freq=1)
+        self.assertResultContains(result,
+                {u'hits': {u'hits': [{u'_score': 0.30685282, u'_type': u'test-type', u'_id': u'4', u'_source': {u'sport': u'football', u'name': u'Cam'}, u'_index': u'test-index'}], u'total': 1, u'max_score': 0.30685282}})
+
 
 class DangerousOperationTests(ElasticSearchTestCase):
     """
     Tests that confirm callers can't do dangerous operations by accident and
     that the substitute routines work
     """
-
     def test_delete_all(self):
         """Make sure ``delete_all()`` sends the right request."""
-        with patch.object(self.conn, '_send_request') as _send_request:
+        with patch.object(self.conn, 'send_request') as send_request:
             self.conn.delete_all('test-index', 'tweet')
-        _send_request.assert_called_once_with(
+        send_request.assert_called_once_with(
             'DELETE',
             ['test-index', 'tweet'],
             query_params={})
@@ -303,9 +420,9 @@ class DangerousOperationTests(ElasticSearchTestCase):
 
     def test_delete_all_indexes(self):
         """Make sure ``delete_all_indexes()`` sends the right request."""
-        with patch.object(self.conn, '_send_request') as _send_request:
+        with patch.object(self.conn, 'send_request') as send_request:
             self.conn.delete_all_indexes()
-        _send_request.assert_called_once_with('DELETE', [''], query_params={})
+        send_request.assert_called_once_with('DELETE', [''], query_params={})
 
     def update_settings_no_args(self):
         """
@@ -316,9 +433,9 @@ class DangerousOperationTests(ElasticSearchTestCase):
 
     def update_all_settings(self):
         """Make sure ``update_all_settings()`` sends the right request."""
-        with patch.object(self.conn, '_send_request') as _send_request:
+        with patch.object(self.conn, 'send_request') as send_request:
             self.conn.update_all_settings({'joe': 'bob'})
-        _send_request.assert_called_once_with(
+        send_request.assert_called_once_with(
             'PUT', ['_settings'], body={'joe': 'bob'})
 
 
@@ -379,7 +496,7 @@ class DowntimePoolingTests(unittest.TestCase):
             session_get.side_effect = Timeout
 
             # This should kill off both servers:
-            for x in xrange(2):
+            for x in range(2):
                 try:
                     conn.get('test-index', 'test-type', 7)
                 except Timeout:
@@ -416,7 +533,6 @@ class KwargsForQueryTests(unittest.TestCase):
         """Test the thing that translates objects to query string text."""
         to_query = ElasticSearch._to_query
         self.assertEqual(to_query(4), '4')
-        self.assertEqual(to_query(4L), '4')
         self.assertEqual(to_query(4.5), '4.5')
         self.assertEqual(to_query(True), 'true')
         self.assertEqual(to_query(('4', 'hi', 'thomas')), '4,hi,thomas')
@@ -426,6 +542,11 @@ class KwargsForQueryTests(unittest.TestCase):
                          '2000-01-02T00:00:00')
         self.assertRaises(TypeError, to_query, object())
 
+        # do not use unittest.skipIf because of python 2.6
+        if not six.PY3:
+            self.assertEqual(to_query(long(4)), '4')
+
+
     def test_es_kwargs(self):
         """
         Make sure ``es_kwargs`` bundles es_ and specifically called out kwargs
@@ -449,7 +570,7 @@ class KwargsForQueryTests(unittest.TestCase):
         def valid_responder(*args, **kwargs):
             """Return an arbitrary successful Response."""
             response = requests.Response()
-            response._content = '{"some": "json"}'
+            response._content = six.b('{"some": "json"}')
             response.status_code = 200
             return response
 
@@ -527,5 +648,41 @@ class KwargsForQueryTests(unittest.TestCase):
         """)
 
 
+class JsonTests(ElasticSearchTestCase):
+    """Tests for JSON encoding and decoding"""
+
+    def test_decimal_encoding(self):
+        """Make sure we can encode ``Decimal`` objects and that they don't end
+        up with quotes around them, which would suggest to ES to represent them
+        as strings if inferring a mapping."""
+        ones = '1.111111111111111111'
+        self.assertEqual(self.conn._encode_json({'hi': Decimal(ones)}),
+                         '{"hi": %s}' % ones)
+
+    def test_set_encoding(self):
+        """Make sure encountering a set doesn't raise a circular reference
+        error."""
+        self.assertEqual(self.conn._encode_json({'hi': set([1])}),
+                         '{"hi": [1]}')
+
+    def test_tuple_encoding(self):
+        """Make sure tuples encode as lists."""
+        self.assertEqual(self.conn._encode_json({'hi': (1, 2, 3)}),
+                         '{"hi": [1, 2, 3]}')
+
+    def test_encoding(self):
+        """Test encoding a zillion other types."""
+        self.assertEqual(self.conn.from_python('abc'), u'abc')
+        self.assertEqual(self.conn.from_python(u'☃'), u'☃')
+        self.assertEqual(self.conn.from_python(123), 123)
+        self.assertEqual(self.conn.from_python(12.2), 12.2)
+        self.assertEqual(self.conn.from_python(True), True)
+        self.assertEqual(self.conn.from_python(False), False)
+        self.assertEqual(self.conn.from_python(date(2011, 12, 30)), '2011-12-30T00:00:00')
+        self.assertEqual(self.conn.from_python(datetime(2011, 12, 30, 11, 59, 32)), '2011-12-30T11:59:32')
+        self.assertEqual(self.conn.from_python([1, 2, 3]), [1, 2, 3])
+        self.assertEqual(self.conn.from_python({'a': 1, 'b': 3, 'c': 2}), {'a': 1, 'b': 3, 'c': 2})
+
+
 if __name__ == '__main__':
     unittest.main()
